#include "SDL_V8.h"

#include "graphic.h"
#include "openscript.h"
#include "opengame.h"
#include "key.h"
#include "mouse.h"
#include "font.h"
#include "mapengine.h"
#include "scriptfs.h"
#include "windowstyle.h"
#include "sound.h"
#include <cstring>
#include "TS.h"

void exitContext(v8::Persistent<v8::Context> context){
    context->v8::Context::Exit();
}

void TS_CallFunc(const char *name, v8::Persistent<v8::Context> context){
        v8::Handle<v8::Value> gameValue = context->Global()->Get(v8::String::New(name));
        if(false){
            printf("TS_CallGame error: game is not defined.");
            return;
        }
        if(!gameValue->IsFunction()){
            printf("TS_CallGame error: game is not a function.");
            return;
        }

        v8::Handle<v8::Function> gameFunc = v8::Handle<v8::Function>::Cast(gameValue);
        gameFunc->Call(context->Global(), 0, NULL);
}

v8::Handle<v8::Value> GetScreenWidth(const v8::Arguments& args)
{
	return  v8::Number::New(SCREENWIDTH);
}

v8::Handle<v8::Value> GetVersionNumber(const v8::Arguments& args)
{
    return v8::Number::New(2.0);
}

v8::Handle<v8::Value> GetVersionString(const v8::Arguments& args)
{
    return v8::String::New(VERSION);
}

v8::Handle<v8::Value> GetScreenHeight(const v8::Arguments& args)
{
	return  v8::Number::New(SCREENHEIGHT);
}


v8::Handle<v8::Value> SetResolution(const v8::Arguments& args)
{
	if(args.Length()<2){
        return v8::ThrowException(v8::String::New("SetResolution Error: Called with fewer than 2 arguments."));
	}

    CHECK_ARG_INT(0, "SetResolution Error: Arg 0 is not a number.");
    CHECK_ARG_INT(1, "SetResolution Error: Arg 1 is not a number.");

    int w = args[0]->IntegerValue();
    int h = args[1]->IntegerValue();

    if(w<MIN_SCREENWIDTH||h<MIN_SCREENHEIGHT){
        return v8::ThrowException(v8::String::New("SetResolution Error: Resolution given too small."));
    }

	SDL_SetVideoMode(w, h, DEPTH, SDL_VIDEO_FLAGS);

	SCREENWIDTH     = w;
	SCREENHEIGHT    = h;
	cmpltscreen.w   = w;
	cmpltscreen.h   = h;
	return v8::True();
}

v8::Handle<v8::Value> GetTime(const v8::Arguments& args)
{
	return  v8::Number::New(SDL_GetTicks());
}



v8::Handle<v8::Value> FlipScreen(const v8::Arguments& args)
{
	//if(!v8::V8::IdleNotification(10)){
	//    printf("Memory Warning\n");
	//    if(!v8::V8::IdleNotification(1000)){
    //        printf("Low Memory\n");
    //        v8::V8::LowMemoryNotification();
    //    }
	//}
	//SDL_WaitThread(EventThrd, NULL);
	//TS_FlipScreen();

    TendEvents();

    SDL_Flip(SDL_GetVideoSurface());
	//SDL_UpdateRect(screen, 0, 0, SCREENWIDTH, SCREENHEIGHT);

	SDL_FillRect(screen, NULL, 0x000000);
	//SDL_Delay(0);
	return v8::Undefined();
}

v8::Handle<v8::Value> SetTitle(const v8::Arguments& args)
{

    if(!SDL_GetVideoInfo()->wm_available) return v8::Undefined();

    CHECK_ARG_STR(0, "SetTitle Error: argument 0 is not a string.");
    v8::String::Utf8Value str(args[0]);
    const char *title = *str;
    SDL_WM_SetCaption(title, NULL );
	return v8::Undefined();
}

v8::Handle<v8::Value> Delay(const v8::Arguments& args)
{
	unsigned int t = args[0]->Uint32Value();
	SDL_Delay(t);
	return v8::Undefined();
}


#ifdef _WIN32
int wmain(int argc, char* argv[]) {
#else
int main(int argc, char* argv[]) {
#endif
	
	v8::Persistent<v8::Context> context = v8::Context::New();
	v8::Persistent<v8::ObjectTemplate> global;

    bool displayinfo = false;
    bool verbosestart= true;
    bool nogame      = false;

    for(int i = 0; i<argc; i++){
        if(strcmp(argv[i], "version")==0){
            printf("TurboSphere Engine %s [Dec 30 2012]\n", VERSION);
            return 0;
        }

        if(strcmp(argv[i], "getinfo")==0){
            displayinfo = true;
            verbosestart= false;
            nogame      = true;
        }


    }
    if(!nogame){
		

        if(argc>1){
            TS_dirs.root = argv[1];
            TS_dirs.root.append("/");
            setDirectories(TS_dirs.root);
        }
        else{
                TS_dirs.root = "startup/";
                setDirectories(TS_dirs.root);
            }
        setConfig(TS_dirs.root);

        opengame(string(TS_dirs.root).append("game.sgm").c_str());
    }

	if (TTF_Init() < 0 ) {
        if(!displayinfo){
        printf("Unable to init SDL ttf: %s\n", TTF_GetError());
        return 1;
        }
    }
       if ( SDL_Init(SDL_FLAGS) < 0 ) {
        if(!displayinfo){
            printf("Unable to init SDL: %s\n", SDL_GetError());
            return 1;
        }
	   }
	   else
	   {
		   SDL_SetClipRect(screen, &cmpltscreen);
        }

	   if (IMG_Init(IMG_FLAGS) <=0) {
        if(!displayinfo){
            printf("Unable to init SDL Image:");
            printf(" %s\n", IMG_GetError());
            return 1;
        }
	   }

    int mix_initted = Mix_Init(MIX_FLAGS);
    if((mix_initted<0)||(mix_initted!=MIX_FLAGS)){
        if(!displayinfo){
            printf("SDL Mixer failed to intialize any loaders. WAV support may still work.\n");
            printf("SDL Mixer error: %s\n", Mix_GetError());
        }
    }

    int audio_rate = 0;
    Uint16 audio_format = 0;
    int audio_channels = 0;

    Mix_QuerySpec(&audio_rate, &audio_format, &audio_channels);

    Mix_OpenAudio(44050, audio_format, audio_channels, 1024);

    //printf("Audio Initialized. %s\n", SDL_GetError());

    if((screen = SDL_SetVideoMode(SCREENWIDTH, SCREENHEIGHT, DEPTH, SDL_VIDEO_FLAGS|((TS_conf.fullscreen)?SDL_FULLSCREEN:0)))==NULL)
    {
		if(!displayinfo){
            printf("SDL Video mode setup failed. Graphics subsystem closing.\n%s\n", SDL_GetError());
            SDL_Quit();
            return 1;
		}
    }
	else
	{
		if(verbosestart) printf("SDL Video mode setup succeeded.\n");
	}

	SDL_ShowCursor(SDL_DISABLE);

    if(displayinfo){

            printf("TurboSphere Engine Test:\n\n");
            PrintSDLInfo();
            return 0;

    }

	std::string ScriptFileText = openfile(TS_conf.mainscript.c_str());

	const char *script_name = TS_conf.gamename.c_str();

	//printf("Script Loaded\n");
	v8::HandleScope handle_scope2;

    atexit(QuitAll);

    //prepare all JS object prototypes.s
    SurfaceInit();
    ImageInit();
    KeyInit();
    BMPFontInit();
    FontInit();
    ColorInit();
    WindowStyleInit();
    FileInit();
    SoundInit();

	v8::Handle<v8::FunctionTemplate> E_Testtempl = v8::FunctionTemplate::New(TS_Test);

	v8::Handle<v8::FunctionTemplate> E_RequireScripttempl = v8::FunctionTemplate::New(TS_LoadScript);

	v8::Handle<v8::FunctionTemplate> E_GetVersionNumbertempl = v8::FunctionTemplate::New(GetVersionNumber);

	v8::Handle<v8::FunctionTemplate> E_GetVersionStringtempl = v8::FunctionTemplate::New(GetVersionString);

	v8::Handle<v8::FunctionTemplate> E_GetScreenWidthtempl = v8::FunctionTemplate::New(GetScreenWidth);

	v8::Handle<v8::FunctionTemplate> E_GetScreenHeighttempl = v8::FunctionTemplate::New(GetScreenHeight);

	v8::Handle<v8::FunctionTemplate> E_SetResolutiontempl = v8::FunctionTemplate::New(SetResolution);

	v8::Handle<v8::FunctionTemplate> E_GetTimetempl = v8::FunctionTemplate::New(GetTime);

	v8::Handle<v8::FunctionTemplate> E_FlipScreentempl = v8::FunctionTemplate::New(FlipScreen);

	v8::Handle<v8::FunctionTemplate> E_GetSystemFonttempl = v8::FunctionTemplate::New(LoadSystemBMPFont);

	v8::Handle<v8::FunctionTemplate> E_Rectangletempl = v8::FunctionTemplate::New(Rectangle);

	v8::Handle<v8::FunctionTemplate> E_GradientRectangletempl = v8::FunctionTemplate::New(GradientRectangle);

	v8::Handle<v8::FunctionTemplate> E_FilledCircletempl = v8::FunctionTemplate::New(FilledCircle);

	v8::Handle<v8::FunctionTemplate> E_OutlinedCircletempl = v8::FunctionTemplate::New(OutlinedCircle);

	v8::Handle<v8::FunctionTemplate> E_Delaytempl = v8::FunctionTemplate::New(Delay);

	v8::Handle<v8::FunctionTemplate> E_SetTitletempl = v8::FunctionTemplate::New(SetTitle);

	v8::Handle<v8::FunctionTemplate> E_Pointtempl = v8::FunctionTemplate::New(Point);

	v8::Handle<v8::FunctionTemplate> E_Linetempl = v8::FunctionTemplate::New(Line);

	v8::Handle<v8::FunctionTemplate> E_Colortempl = v8::FunctionTemplate::New(CreateColor);

	v8::Handle<v8::FunctionTemplate> E_TTFFonttempl = v8::FunctionTemplate::New(LoadTTFFont);

	v8::Handle<v8::FunctionTemplate> E_BMPFonttempl = v8::FunctionTemplate::New(LoadBMPFont);

	v8::Handle<v8::FunctionTemplate> E_Surfacetempl = v8::FunctionTemplate::New(TS_CreateSurface);

	v8::Handle<v8::FunctionTemplate> E_Imagetempl = v8::FunctionTemplate::New(TS_CreateImage);

	v8::Handle<v8::FunctionTemplate> E_SysArrowtempl = v8::FunctionTemplate::New(TS_GetSystemArrow);

	v8::Handle<v8::FunctionTemplate> E_KeyConsttempl = v8::FunctionTemplate::New();

	v8::Handle<v8::FunctionTemplate> E_Filetempl = v8::FunctionTemplate::New(TS_OpenFile);

	v8::Handle<v8::FunctionTemplate> E_WindowStyletempl = v8::FunctionTemplate::New(LoadWindowStyle);

	v8::Handle<v8::FunctionTemplate> E_SysWindowStyletempl = v8::FunctionTemplate::New(GetSystemWindowStyle);

	v8::Handle<v8::FunctionTemplate> E_GetKeytempl = v8::FunctionTemplate::New(GetKey);

	v8::Handle<v8::FunctionTemplate> E_AreKeysLefttempl = v8::FunctionTemplate::New(AreKeysLeft);

	v8::Handle<v8::FunctionTemplate> E_IsKeyPressedtempl = v8::FunctionTemplate::New(IsKeyPressed);

	v8::Handle<v8::FunctionTemplate> E_IsAnyKeyPressedtempl = v8::FunctionTemplate::New(IsAnyKeyPressed);

	v8::Handle<v8::FunctionTemplate> E_GetKeyStringtempl = v8::FunctionTemplate::New(GetKeyString);

	v8::Handle<v8::FunctionTemplate> E_GetMouseXtempl = v8::FunctionTemplate::New(GetMouseX);

	v8::Handle<v8::FunctionTemplate> E_GetMouseYtempl = v8::FunctionTemplate::New(GetMouseY);

	v8::Handle<v8::FunctionTemplate> E_IsMouseButtonPressedtempl = v8::FunctionTemplate::New(IsMouseButtonPressed);

	v8::Handle<v8::FunctionTemplate> E_MapEnginetempl = v8::FunctionTemplate::New(MapEngine);

	v8::Handle<v8::FunctionTemplate> E_ExitMapEnginetempl = v8::FunctionTemplate::New(ExitMapEngine);

	v8::Handle<v8::FunctionTemplate> E_IsMapEngineRunningtempl = v8::FunctionTemplate::New(IsMapEngineRunning);

	v8::Handle<v8::FunctionTemplate> E_Soundtempl = v8::FunctionTemplate::New(LoadSound);

	//v8::Handle<v8::Object> KEY_ENTER = KeyConstInsttempl->NewInstance();

	FontInsttempl->SetAccessor(v8::String::New("fontName"), GetFontname, SetFontname);
	FontInsttempl->SetAccessor(v8::String::New("size"), GetFontsize, SetFontsize);
	FontInsttempl->SetAccessor(v8::String::New("colorMask"), GetFontMask, SetFontMask);


	//FontInsttempl->Set("drawText", v8::FunctionTemplate::New(TS_drawText));

	ColorInsttempl->SetAccessor(v8::String::New("red"), GetRed, SetRed);
	ColorInsttempl->SetAccessor(v8::String::New("green"), GetGreen, SetGreen);
	ColorInsttempl->SetAccessor(v8::String::New("blue"), GetBlue, SetBlue);
	ColorInsttempl->SetAccessor(v8::String::New("alpha"), GetAlpha, SetAlpha);

	SurfaceInsttempl->SetAccessor(v8::String::New("width"), GetWidth, SetWidth);
	SurfaceInsttempl->SetAccessor(v8::String::New("height"), GetHeight, SetHeight);
	ImageInsttempl->SetAccessor(v8::String::New("width"), GetWidth, SetWidth);
	ImageInsttempl->SetAccessor(v8::String::New("height"),  GetHeight, SetHeight);

    v8::Context::Scope context_scope(context);

	context->Global()->Set(v8::String::New("TS_Test"), E_Testtempl->GetFunction());

	context->Global()->Set(v8::String::New("RequireScript"), E_RequireScripttempl->GetFunction());

	context->Global()->Set(v8::String::New("GetVersion"), E_GetVersionNumbertempl->GetFunction());

	context->Global()->Set(v8::String::New("GetVersionString"), E_GetVersionStringtempl->GetFunction());

	context->Global()->Set(v8::String::New("GetScreenHeight"), E_GetScreenHeighttempl->GetFunction());

	context->Global()->Set(v8::String::New("GetScreenWidth"), E_GetScreenWidthtempl->GetFunction());

	context->Global()->Set(v8::String::New("SetResolution"), E_SetResolutiontempl->GetFunction());

	context->Global()->Set(v8::String::New("GetTime"), E_GetTimetempl->GetFunction());

	context->Global()->Set(v8::String::New("FlipScreen"), E_FlipScreentempl->GetFunction());

	context->Global()->Set(v8::String::New("GetSystemFont"), E_GetSystemFonttempl->GetFunction());

	context->Global()->Set(v8::String::New("Rectangle"), E_Rectangletempl->GetFunction());

	context->Global()->Set(v8::String::New("GradientRectangle"), E_GradientRectangletempl->GetFunction());

	context->Global()->Set(v8::String::New("FilledCircle"), E_FilledCircletempl->GetFunction());

	context->Global()->Set(v8::String::New("OutlinedCircle"), E_OutlinedCircletempl->GetFunction());

	context->Global()->Set(v8::String::New("Delay"), E_Delaytempl->GetFunction());

	context->Global()->Set(v8::String::New("SetTitle"), E_SetTitletempl->GetFunction());

	context->Global()->Set(v8::String::New("Point"), E_Pointtempl->GetFunction());

	context->Global()->Set(v8::String::New("Line"), E_Linetempl->GetFunction());

	context->Global()->Set(v8::String::New("Color"), E_Colortempl->GetFunction());

	context->Global()->Set(v8::String::New("TTFFont"), E_TTFFonttempl->GetFunction());

	context->Global()->Set(v8::String::New("Font"), E_BMPFonttempl->GetFunction());

	context->Global()->Set(v8::String::New("Surface"), E_Surfacetempl->GetFunction());

	context->Global()->Set(v8::String::New("Image"), E_Imagetempl->GetFunction());

	context->Global()->Set(v8::String::New("GetSystemArrow"), E_SysArrowtempl->GetFunction());

	context->Global()->Set(v8::String::New("KEY_CONST"), E_KeyConsttempl->GetFunction());

	context->Global()->Set(v8::String::New("File"), E_Filetempl->GetFunction());

	context->Global()->Set(v8::String::New("WindowStyle"), E_WindowStyletempl->GetFunction());

	context->Global()->Set(v8::String::New("GetSystemWindowStyle"), E_SysWindowStyletempl->GetFunction());

	context->Global()->Set(v8::String::New("GetKey"), E_GetKeytempl->GetFunction());

	context->Global()->Set(v8::String::New("AreKeysLeft"), E_AreKeysLefttempl->GetFunction());

	context->Global()->Set(v8::String::New("IsKeyPressed"), E_IsKeyPressedtempl->GetFunction());

	context->Global()->Set(v8::String::New("IsAnyKeyPressed"), E_IsAnyKeyPressedtempl->GetFunction());

	context->Global()->Set(v8::String::New("GetKeyString"), E_GetKeyStringtempl->GetFunction());

	context->Global()->Set(v8::String::New("GetMouseX"), E_GetMouseXtempl->GetFunction());

	context->Global()->Set(v8::String::New("GetMouseY"), E_GetMouseYtempl->GetFunction());

	context->Global()->Set(v8::String::New("IsMouseButtonPressed"), E_IsMouseButtonPressedtempl->GetFunction());

	context->Global()->Set(v8::String::New("MapEngine"), E_MapEnginetempl->GetFunction());

	context->Global()->Set(v8::String::New("ExitMapEngine"), E_ExitMapEnginetempl->GetFunction());

	context->Global()->Set(v8::String::New("IsMapEngineRunning"), E_IsMapEngineRunningtempl->GetFunction());

	context->Global()->Set(v8::String::New("Sound"), E_Soundtempl->GetFunction());

	TS_SetKeyConstants(context);

	TS_SetMouseConstants(context);

    int start = SDL_GetTicks();

    //EventHandlerInit();

    //SDL_WaitThread(EventThrd, NULL);

	ExecuteString(v8::String::New(ScriptFileText.c_str()), v8::String::New(script_name), true);

    TS_CallFunc("game", context);


	//printf("Executed JS source\n");

	int end = SDL_GetTicks();

	unsigned int time = end-start;
std::stringstream timestr;
	timestr.str("\n");
	timestr<<start;
	timestr<<", ";
	timestr<<end;
	timestr<<", ";
	timestr<<time;
std::string STR_timestr = timestr.str();
const char* CSTR_timestr= STR_timestr.c_str();
  printf(CSTR_timestr);
  printf("\nV8 execution finished");
  printf("\n");

  SDL_VideoQuit();

  //Not necessary, but is it really the OS's resposibility to clean up after programs?
  v8::V8::LowMemoryNotification();

    //QuitAll();
  return 0;
}
